angular.module('mnQuery').controller('mnQueryController',
    function ($scope, $modal, $timeout, mnQueryService, mnHelper, ModalService) {

  //
  // automatic result limiter
  //
  
  $scope.limit = mnQueryService.limit;
  
  $scope.version = "1.0.0";
  
  //
  // metadata
  //

  $scope.buckets = mnQueryService.buckets;
  $scope.busyGettingBuckets = mnQueryService.busyGettingBuckets;
  $scope.updateBuckets = mnQueryService.updateBuckets;

  //
  // variable and code for managing the choice of output format
  //

  $scope.selectTab = function(tabNum) {
    mnQueryService.selectTab(tabNum);

    // select focus after a delay to try and force update of the editor
    $timeout(function(){
      if ($scope.outputEditor) {
        $scope.outputEditor.focus();
        $scope.inputEditor.focus();
      }
    },10);
    
  };
  $scope.isSelected = mnQueryService.isSelected;

  
  //
  // structure for holding the current query and result
  //    

  $scope.lastResult = mnQueryService.getResult(); 
  $scope.test = "{\"foo\" : \"bar\"}";

  //
  // options for the two editors, query and result
  //

  $scope.aceInputLoaded = function(_editor) {
    var langTools = ace.require("ace/ext/language_tools");

    _editor.$blockScrolling = Infinity;
    _editor.setOptions({enableBasicAutocompletion: true});
    $scope.inputEditor = _editor;
    
    var identifierCompleter = {
      getCompletions: function(editor, session, pos, prefix, callback) {
        callback(null,mnQueryService.autoCompleteArray);
      }  
    };
    langTools.addCompleter(identifierCompleter);
  };

  
  var endsWithSemi = /;\s*$/i;

  $scope.aceInputChanged = function(e) {
    var curText = $scope.inputEditor.getSession().getValue(); 
    if (curText == 'Enter a query here...') {
      $scope.inputEditor.selectAll();
      $scope.inputEditor.focus();
    }
    $scope.inputEditor.$blockScrolling = Infinity;
    
    // if they hit enter and the query ends with a semicolon, run the query

    if (e[0].action === 'insert' && e[0].end.column === 0 && 
        e[0].start.row+1 == e[0].end.row && e[0].start.column > 0 &&
        endsWithSemi.test($scope.lastResult.query))
      $scope.query();
  };

  // to keep Ace happy, need to set the following parameter on load
  $scope.aceLoaded = function(_editor) {
    //console.log("AceOutputLoaded");
    _editor.$blockScrolling = Infinity;
    _editor.setReadOnly(true);
    $scope.outputEditor = _editor;
    var editor_size = ($('#whole_ui').height() - $('#query_box').height() - $('#result_header').height() - $('#north_ui').height()) - 76; 
    $('#result_editor').height(editor_size);
  };

  // when we get more data, make sure the editor is the correct size
  $scope.aceChanged = function(e) {
    //console.log("AceInputLoaded");
    var editor_size = ($('#whole_ui').height() - $('#query_box').height() - $('#result_header').height() - $('#north_ui').height()) - 76; 
    $('#result_editor').height(editor_size);
    if ($scope.outputEditor)
      $scope.outputEditor.$blockScrolling = Infinity;
  };

  $scope.aceOptions = {
      mode: 'json',
      showGutter: true,
      onLoad: $scope.aceLoaded,
      onChange: $scope.aceChanged,
      $blockScrolling: Infinity
  };

  $scope.aceInputOptions = {
      mode: 'n1ql',
      showGutter: true,
      onLoad: $scope.aceInputLoaded,
      onChange: $scope.aceInputChanged,
      $blockScrolling: Infinity
  };


  /*
   * Whenever the query ends with ";\n", issue the query. This is a shortcut to avoid 
   * finding the mouse and clicking "Execute".
   */
//  $scope.$watch("lastResult.query",function(val) {
//    var endsWithSemi = /;\s*$/i;
//    if ($scope.lastResult.query.length > 2 && 
//        endsWithSemi.test($scope.lastResult.query))
//      $scope.query();
//  });

  //
  // some functions for handling query history, going backward and forward
  //

  $scope.prev = mnQueryService.prevResult;
  $scope.next = mnQueryService.nextResult;

  $scope.hasNext = mnQueryService.hasNextResult;
  $scope.hasPrev = mnQueryService.hasPrevResult;

  //
  // functions for running queries and explain
  //

  $scope.busyExecutingQuery = mnQueryService.busyExecutingQuery;

  $scope.focusOnInput = function()  {
    if ($scope.inputEditor && $scope.inputEditor.focus)
      $scope.inputEditor.focus();    
  }  

  $scope.query = function() {
    var queryStr = $scope.lastResult.query;
    var hasLimit = /limit\s+\d+\s*;\s*$/i;
    var endsWithSemi = /;\s*$/i;
    var startsWithSelect = /^\s*select/i;
    
//    console.log("Got query: " + queryStr);
//    console.log("endsWithSemi: " + endsWithSemi.test(queryStr));
//    console.log("startsWithSelect: " + startsWithSelect.test(queryStr));
    
    // add a limit to all "select" statements by wrapping
    if (startsWithSelect.test(queryStr)) {
      // strip out any ending semicolon
      if (endsWithSemi.test(queryStr))
        queryStr = queryStr.replace(endsWithSemi,"");
      
      // handle garbage in the limit dialog
      if (isNaN(Number(mnQueryService.limit.max)))
        mnQueryService.limit.max = 50;

      // wrap the query in a new query with a limit 
      queryStr = "select xyzzyplugh.* from (" + queryStr + ") xyzzyplugh limit " + mnQueryService.limit.max + ";";
    }

    //  console.log("New Query: " + queryStr);

    // run the query and show a spinner

    var promise = mnQueryService.executeQuery(queryStr);

    if (promise) {
      // for long queries, show a spinner
      mnHelper.promiseHelper($scope,promise/*, dialog*/).showSpinner("queryInProgress");
      // also have the input grab focus at the end
      promise.success($scope.focusOnInput)
      .error($scope.focusOnInput);
    }
    
    
    //helper.closeOnSuccess();
  };

  $scope.file = {name: "data.json"};
  
  $scope.save = function() {
    var isSafari = /^((?!chrome).)*safari/i.test(navigator.userAgent);
    
    // safari does'nt support saveAs
    if (isSafari) {
      var file = new Blob([$scope.lastResult.result],{type: "text/json"});
      saveAs(file,$scope.file.name);      
      return;
    }


    // get a name for the file
    var promise = $modal.open({
      templateUrl: 'query/file_dialog/mn_query_file_dialog.html',
      scope: $scope              
    }).result;
    
    // now save it
    promise.then(function (res) {
      console.log("Promise, file: " + $scope.file.name + ", res: " + res);
      var file = new Blob([$scope.lastResult.result],{type: "text/json"});
      saveAs(file,$scope.file.name);
    });
 };

  // let's start off with a list of the buckets
  $scope.updateBuckets();
});


