angular.module('mnQueryService').factory('mnQueryService',
    function ($q, $timeout, $http, $modal, mnHttp, mnHelper, ModalService) {

  var mnQueryService = {};
  //
  // remember which tab is selected for output style: JSON, table, or tree
  //

  var outputTab = 1;
  mnQueryService.selectTab = function(newTab) {outputTab = newTab;};
  mnQueryService.isSelected = function(checkTab) {return outputTab === checkTab;};

  mnQueryService.limit = {max: 50};
  
  //
  // keep track of the bucket and field names we have seen, for use in autocompletion
  //
  
  mnQueryService.autoCompleteTokens = {}; // keep a map, name and kind
  mnQueryService.autoCompleteArray = []; // array for use with Ace Editor  
  
  function addToken(token, type) {
    // see if the token needs to be quoted
    if (token.indexOf(' ') >= 0 || token.indexOf('-') >= 0)
      mnQueryService.autoCompleteTokens['`' + token + '`'] = type;
    else
      mnQueryService.autoCompleteTokens[token] = type;
  };
  
  
  function refreshAutoCompleteArray() {
    mnQueryService.autoCompleteArray.length = 0;
    
    for (var key in mnQueryService.autoCompleteTokens) {
      mnQueryService.autoCompleteArray.push(
          {caption:key,snippet:key,meta:mnQueryService.autoCompleteTokens[key]});
    }
  };
  
  
  //
  // go over a schema and recursively put all the field names in our name map
  //
  
  function getFieldNamesFromSchema(schema) {
    for (var i=0; i< schema.length; i++)
      for (var j=0; j < schema[i]['fields'].length; j++) {
        var field = schema[i]['fields'][j];
        addToken(field['field_name'],"field");

        if (field['subtype'])
          getFieldNamesFromSchema(field['subtype']);
      }
  };
  
  //
  // this structure holds the current query text, the current query result,
  // and defines the object for holding the query history
  //

  function QueryResult(hide,status,elapsedTime,executionTime,resultCount,resultSize,result,data,query,requestID) {
    this.hide = hide;
    this.status = status;
    this.elapsedTime = elapsedTime;
    this.executionTime = executionTime;
    this.resultCount = resultCount;
    this.resultSize = resultSize;
    this.result = result;
    this.data = data;
    this.query = query;
    this.requestID = requestID;
  };
  QueryResult.prototype.clone = function()
  {
    return new QueryResult(this.hide,this.status,this.elapsedTime,this.executionTime,this.resultCount,
        this.resultSize,this.result,this.data,this.query,this.requestID);
  };
  QueryResult.prototype.copyIn = function(other)
  {
    this.hide = other.hide;
    this.status = other.status;
    this.elapsedTime = other.elapsedTime;
    this.executionTime = other.executionTime;
    this.resultCount = other.resultCount;
    this.resultSize = other.resultSize;
    this.result = other.result;
    this.data = other.data;
    this.query = other.query;
    this.requestID = other.requestID;
  };

  var lastResult = new QueryResult(false,'N/A','N/A','N/A','N/A','N/A','',{},'Enter a query here...');
  mnQueryService.getResult = function() {return lastResult;};
  
  var dummyResult = new QueryResult(false,'dummy','dummy','dummy','dummy','dummy','',{},'dummy');

  var pastQueries = [];       // keep a history of past queries and their results 
  var currentQueryIndex = -1; // where in the array are we?

  //
  // we also keep a history of executed queries and their results
  // we will permit forward and backward traversal of the history
  //
  mnQueryService.hasPrevResult = function() {return currentQueryIndex > 0};
  mnQueryService.hasNextResult = function() {return currentQueryIndex < pastQueries.length-1};

  mnQueryService.prevResult = function()
  {
    if (currentQueryIndex > 0) // can't go earlier than the 1st
    {
      // 
      // the following gross hack is due to an angular issue where it doesn't 
      // successfully detect *some* changes in the result data, and thus doesn't 
      // update the table. So we set the result to blank, then back to a value 
      // after a certain delay.
      //
      
      lastResult.result = "";
      lastResult.data = "{}";

      $timeout(function(){
        currentQueryIndex--;
        lastResult.copyIn(pastQueries[currentQueryIndex]);
        currentQuery = lastResult.query;        
      },50);
      //return(currentQuery);
    }
  }

  mnQueryService.nextResult = function()
  {
    if (currentQueryIndex < pastQueries.length -1) // can we go forward?
    {
      
      // 
      // see comment above about the delay hack.
      //
      
      lastResult.result = "";
      lastResult.data = "{}";

      $timeout(function(){
        currentQueryIndex++;
        lastResult.copyIn(pastQueries[currentQueryIndex]);
        currentQuery = lastResult.query;
      },50);
      //return(currentQuery);
    }
  }

  //
  // call the proxy which handles N1QL queries
  //

  mnQueryService.busyExecutingQuery = false;
  
  mnQueryService.executeQuery = function(queryText) {

    // don't allow multiple queries, as indicated by anything after a semicolon

    var stuffAfterSemi = /;\s*\S+/i;
    if (stuffAfterSemi.test(queryText)) {
      result = '{"error": "you cannot isue more than one query at once."}';
      data = {error: "Error, you cannot isue more than one query at once."};
      newResult = new QueryResult(false,"error",0.0,0.0,0,'0B',
          result,data,lastResult.query);
      lastResult.copyIn(newResult);
      pastQueries.push(newResult);
      currentQueryIndex = pastQueries.length - 1; // after run, set current result to end
      return null;
    }


    // don't allow multiple queries at once
    if (mnQueryService.busyExecutingQuery)
      return;
    
    mnQueryService.busyExecutingQuery = true;
    lastResult.result = '{"status": "Executing Query"}';
    lastResult.data = "{}";

//    var pre_post_ms = new Date().getTime();
    
    //
    // create a data structure for holding the query, and the credentials for any SASL
    // protected buckets
    //
    
    var queryData = {statement: queryText};
    var credString = "";
    for (var i = 0; i < mnQueryService.buckets.length; i++) { 
        if (credString.length > 0)
          credString += ",";
        
        var pw = mnQueryService.buckets[i].password ? 
            mnQueryService.buckets[i].password : "";

            credString += '{"user":"local:'+mnQueryService.buckets[i].id+'","pass":"' + 
            pw +'"}';
      }
    
    if (credString.length > 0)
      queryData.creds = '[' + credString + ']';

    // send the query off via REST API
    return mnHttp.post("/query/service",queryData)
    .success(function(data, status, headers, config) {
      var result;
      
//      var post_post_ms = new Date().getTime();

      // make sure we show errors if the query did not succeed
      if (data.status == "success")
        result = data.results;
      else {
        var failed = "Authorization Failed";
        // hack - detect authorization failed, make a suggestion
        for (var i=0; i < data.errors.length; i++)
        if (data.errors[i].msg && 
            data.errors[i].msg.length >= failed.length &&
            data.errors[i].msg.substring(0,failed.length) == failed)
          data.errors[i].suggestion = "Try authorizing the necessary bucket(s) in the metadata panel to the left.";

        result = data.errors;
      }
      
      // if we got no metrics, create a dummy version
      
      if (!data.metrics) {
        data.metrics = {elapsedTime: 0.0, executionTime: 0.0, resultCount: 0, resultSize: "0B", elapsedTime: 0.0}
      }
      
      newResult = new QueryResult(false,data.status,data.metrics.elapsedTime,
          data.metrics.executionTime,data.metrics.resultCount,data.metrics.resultSize + 'B',
          angular.toJson(result, true),result,lastResult.query,data.requestID);
      lastResult.copyIn(newResult);
      pastQueries.push(newResult);
      currentQueryIndex = pastQueries.length - 1; // after run, set current result to end
      
      mnQueryService.busyExecutingQuery = false;      
      
//      var post2_ms = new Date().getTime();

//      $timeout(function(){
//        var post3_ms = new Date().getTime();
//        var diff1 = post_post_ms - pre_post_ms;
//        var diff2 = post2_ms - post_post_ms; 
//        var diff3 = post3_ms - post2_ms; 
//        console.log("Query execution time: " + diff1 + ", processing: " + diff2 + ", rendering: " + diff3);
//      },10);
      
    })
    .error(function(data, status, headers, config) {
        console.log("Error Data: " + JSON.stringify(data));
        if (!data.metrics) {
          data.metrics = {elapsedTime: 0.0, executionTime: 0.0, resultCount: 0, resultSize: "0B", elapsedTime: 0.0}
        }

      newResult = new QueryResult(false,data.status,data.metrics.elapsedTime,
          data.metrics.executionTime,data.metrics.resultCount,data.metrics.resultSize + 'B',
          angular.toJson(data.errors, true),data.errors,lastResult.query,data.requestID);
      lastResult.copyIn(newResult);
      pastQueries.push(newResult);
      currentQueryIndex = pastQueries.length - 1; // after run, set current result to end
      
      mnQueryService.busyExecutingQuery = false;      
    });
  }

  //
  // manage metadata, including buckets, fields, and field descriptions
  //
  
  mnQueryService.buckets = [];
  mnQueryService.busyGettingBuckets = false;  
  mnQueryService.updateBuckets = function() {
    if (mnQueryService.busyGettingBuckets)
      return;
    
    mnQueryService.busyGettingBuckets = true;
    mnQueryService.buckets.length = 0;
    
    // use a query to get buckets with a primary index
    //var queryText = "select distinct indexes.keyspace_id from system:indexes where is_primary = true order by indexes.keyspace_id ;";
    var queryText2 = "select indexes.keyspace_id, array_agg(index_key) primary_indexes " +
      "from system:indexes group by indexes.keyspace_id having array_agg(is_primary) " +
      "order by indexes.keyspace_id;";
    
    var queryText = 
      "select max(keyspace_id) id, max(has_primary) has_prim, max(has_second) has_sec, max(secondary_indexes) sec_ind from (" +
      " select indexes.keyspace_id, true has_primary" + 
      "  from system:indexes where is_primary = true" +
      "  union" +
      "  select indexes.keyspace_id, true has_second, array_agg(indexes.index_key) secondary_indexes, indexes.condition" + 
      "  from system:indexes where is_primary is missing or is_primary = false group by keyspace_id having keyspace_id is not null" +
      "  union" +
      "   select id keyspace_id from system:keyspaces except (select indexes.keyspace_id from system:indexes union select \"\" keyspace_id)" +
      "  ) foo group by keyspace_id having keyspace_id is not null order by keyspace_id";
    
    res1 = mnHttp.post("/query/service",{statement : queryText })
    .success(function(data, status, headers, config) {

      var bucket_names = [];
      var passwords = [];
      
      if (data && data.results) for (var i=0; i< data.results.length; i++) {
        var bucket = data.results[i];
        bucket.expanded = false;
        bucket.schema = [];
        bucket_names.push(bucket.id);
        passwords.push(""); // assume no password for now
        mnQueryService.buckets.push(bucket);
        addToken(bucket.id,"bucket");
      }
      refreshAutoCompleteArray();      
      
      //
      // after the screen refreshes, find out which buckets require passwords
      //
      
      if (bucket_names.length > 0) $timeout(function(){
        mnQueryService.authenticateBuckets(bucket_names,passwords,
            function(data, status, headers, config) {
              if (data && data.success) for (var i=0; i < data.success.length; i++) 
                mnQueryService.buckets[i].passwordNeeded = !data.success[i]; 
            },
            function(data, status, headers, config) {
              console.log(data.errors);
            }
        );
      },10);

            
      mnQueryService.busyGettingBuckets = false;
    })
    .error(function(data, status, headers, config) {
      mnQueryService.buckets.push({id: data.errors, schema: []});
      mnQueryService.busyGettingBuckets = false;
    });
    
  };
  
  //
  //  
  //

  mnQueryService.authenticateBuckets = function(bucket_names, passwords, onSuccess, onError) {
    mnHttp.post("/authenticate",{bucket : bucket_names, password: passwords})
    .success(onSuccess)
    .error(onError);      
  };

  
  //
  // Get a schema for a given, named bucket.
  //
  
  mnQueryService.getSchemaForBucket = function(bucket) {

    var queryText = "describe `" + bucket.id + "`;";
    var queryData = {statement: queryText};
    if (bucket.password)
      queryData.creds = '[{"user":"local:'+bucket.id+'","pass":"' + bucket.password +'"}]';

   return mnHttp.post("/query/service",queryData)
    .success(function(data, status, headers, config) {
      //console.log("Done!");
      bucket.schema.length = 0;
      //console.log("Schema status: " + status);
      //console.log("Schema results: " + JSON.stringify(data));
      if (_.isString(data.results))
        bucket.schema_error = data.results;
      else if (data.results && data.results.error)
        bucket.schema_error = data.results.error;
      else {
        bucket.schema = data.results[0];

        var totalDocCount = 0;
        for (var i=0; i<bucket.schema.length; i++)
          totalDocCount += bucket.schema[i]['#docs'];

        getFieldNamesFromSchema(bucket.schema);        
        refreshAutoCompleteArray();      

        //console.log("for bucket: " + bucket.name + " got doc count: " + totalDocCount)
        bucket.totalDocCount = totalDocCount;

        for (var i=0; i<bucket.schema.length; i++)
          bucket.schema[i]['%docs'] = (bucket.schema[i]['#docs']/totalDocCount*100);
        
        // we have an array of columns that are indexed. Let's mark the individual
        // fields, now that we have a schema. 
        bucket.indexed_fields = {};
        
        // each element of the sec_ind array is an array of field names, turn into a map
        _.forEach(bucket.sec_ind,function(elem) {
          _.forEach(elem,function(field) {
                bucket.indexed_fields[field] = true;
              })});

        for (var flavor=0; flavor<bucket.schema.length; flavor++) // iterate over flavors
          markIndexedFields(bucket.indexed_fields, bucket.schema[flavor], "");
        
        bucket.schema.unshift({Summary: "Summary: " + bucket.schema.length + " document flavors found from a sample of "+ totalDocCount + " documents"});        
      }

    })
    .error(function(data, status, headers, config) {
      console.log(data.errors);
    });
    
  };
  
  //
  // When we get the schema, we need to mark the indexed fields. We start at the top
  // level, but recursively traverse any subtypes, keeping track of the path that we
  // followed to get to the subtype.
  // 
  
  function markIndexedFields(fieldMap, schema, path) {
    //console.log("marking schema size: "+schema.fields.length + " with path: " + path);
      for (var field=0; field<schema.fields.length; field++) {
        var theField = schema.fields[field];
        // in the list of indexed fields, the field names are quoted with back quotes
        var quoted_field_name = '`' + theField.field_name + '`';
        if (path.length > 0)
          quoted_field_name = '(' + path + quoted_field_name + ')';
        
        // are we in the index map?
        if (fieldMap[quoted_field_name]) {
          theField.indexed = true;
        }
        
        // do we have a subtype to traverse?
        if (theField.subtype)
          markIndexedFields(fieldMap,theField.subtype,path + '`' + theField.field_name + '`.');
      }
  };
  
  
  return mnQueryService;
});

